/*
 * videowidget.cpp
 *
 * Copyright (C) 2008 Christophe Thommeret <hftom@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <KDebug>
#include <KStandardDirs>

#include <QApplication>
#include <QMouseEvent>
#include <QLabel>
#include <QDir>

#include "videowidget.h"
#include "osd.h"



VideoWidget::VideoWidget( QWidget *parent, QString mrl ) : QGLWidget( parent )
{
	func = new GLFunc( context() );
	glv = new GLVideo( this, func );

	osd = new Osd( this, func );

	xineMrl = "rtp://192.168.0.255:1234";

	xine = NULL;
	setMouseTracking( true );
	setMinimumSize( 300, 200 );
	connect( glv, SIGNAL(wantRepaint()), this, SLOT(repaintRequest()) );
	connect( this, SIGNAL(newFrame(const int&,const int&,const int&,const double&,void*,void*,void*)),
		glv, SLOT(updateFrame(const int&,const int&,const int&,const double&,void*,void*,void*)), Qt::BlockingQueuedConnection );
	connect( this, SIGNAL(newOverlays(const int&,raw_overlay_t*)),
		glv, SLOT(updateOverlays(const int&,raw_overlay_t*)), Qt::BlockingQueuedConnection );
	connect( &redrawTimer, SIGNAL(timeout()), this, SLOT(checkNeededRepaint()) );
	redrawTimer.start( 10 );
	fps = numFrames = 0;
	connect( &fpsTimer, SIGNAL(timeout()), this, SLOT(setFPS()) );
	fpsTimer.start( 1000 );
	animRepaint = wantRepaint = 0;

	setAutoFillBackground( false ); //overpainting

	xineThumb = new XineThumb();
	xineThumb->setThumbDir( KStandardDirs::locateLocal("data", "kaffeinegl/thumbs/", KGlobal::dirs() ) );
}



VideoWidget::~VideoWidget()
{
	delete xineThumb;
	if ( xine )
		delete xine;
}



void VideoWidget::toggleDeinterlace()
{
	if ( xine )
		xine->toggleDeinterlace();
}



QString VideoWidget::shot()
{
	QString s;
	QImage image;
	if ( glv->takeVideoShot( image, THUMB_WIDTH ) ) {
		QString path = KStandardDirs::locateLocal("data", "kaffeinegl/thumbs/", KGlobal::dirs() );
		s = xineMrl;
		image.save( path+s.remove("/").append(".jpg") );
	}
	return s;
}



void VideoWidget::thumb( const QStringList &sl )
{
	xineThumb->prepend( sl );
}



void VideoWidget::startRepaint()
{
	++animRepaint;
}



void VideoWidget::stopRepaint()
{
	--animRepaint;
	++wantRepaint;
}



void VideoWidget::checkNeededRepaint()
{
	if ( wantRepaint || animRepaint )
		updateGL();
}



void VideoWidget::repaintRequest()
{
	++wantRepaint;
}



void VideoWidget::setFPS()
{
	fps = numFrames;
	numFrames = 0;
	repaintRequest();
}



void VideoWidget::playMrl( QString mrl )
{
	xineMrl = mrl;
	play();
}



void VideoWidget::play()
{
	if ( !xine )
		xine = new XineEngine( (void*)this, glv->supportedVideoFormats() );
	xine->play( xineMrl );
}



void VideoWidget::stop()
{
	if ( xine )
		xine->stop();
}



void VideoWidget::initializeGL()
{
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	glClearDepth( 1.0f );
	glDepthFunc( GL_LEQUAL );
	glEnable( GL_DEPTH_TEST );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glEnable( GL_BLEND );
	glShadeModel( GL_SMOOTH );
	glEnable( GL_TEXTURE_2D );
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glv->initGL();
	if ( func->doVideoSync() )
		func->mglXSwapInterval( 1 );
}



void VideoWidget::resizeGL( int width, int height )
{
	glViewport( 0, 0, width, height );
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum( 0, width, height, 0, 4.9999, 60.0 );
	glMatrixMode(GL_MODELVIEW);
	osd->resize( width, height );
	repaintRequest();
}



void VideoWidget::paintGL()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glLoadIdentity();

	glv->draw( width(), height(), true );
	osd->paint();


	glTranslatef( width()-50, 0, 0 );
	glRotatef( 45, 0, 0, 1 );
	QRect rect( -40,0,180,30 );
	func->drawQuad( rect, QColor(0,0,0,128) );
	glColor4f( 1, 1, 1, 1 );
	renderText( 5, 20, QString("%1 fps").arg(fps), QFont( "Sans", 16 ) );

	++numFrames;
	wantRepaint = 0;
}



void VideoWidget::raw_output_cb( void *user_data, int frame_format, int frame_width,
	int frame_height, double frame_aspect, void *data0, void *data1, void *data2 )
{
	VideoWidget* vw = (VideoWidget*)user_data;
	emit vw->newFrame( frame_format, frame_width, frame_height, frame_aspect, data0, data1, data2 );
}



void VideoWidget::raw_overlay_cb(void *user_data, int num_ovl, raw_overlay_t *overlay_array)
{
	VideoWidget* vw = (VideoWidget*)user_data;
	emit vw->newOverlays( num_ovl, overlay_array );
}



void VideoWidget::wheelEvent( QWheelEvent *e )
{
	osd->wheelEvent( e );
}



void VideoWidget::mouseReleaseEvent( QMouseEvent *mev )
{
	osd->mouseReleased( mev );
}



void VideoWidget::mouseDoubleClickEvent( QMouseEvent *e )
{
	if ( e->button()!=Qt::LeftButton )
		return;
	/*if ( parentWidget()->isFullScreen() ) {
		parentWidget()->showNormal();
		kDebug() << "showNormal()";
	}
	else {
		parentWidget()->showFullScreen();
		kDebug() << "showFullscreen()";
	}*/
	parentWidget()->setWindowState( parentWidget()->windowState() ^ Qt::WindowFullScreen );
}



void VideoWidget::mouseMoveEvent(QMouseEvent* mev)
{
	osd->mouseMove( mev );

	if ( !xine)
		return;

	if ( mev->x()>=glv->videoTopLeft.x() || mev->x()<=glv->videoBottomRight.x() || mev->y()>=glv->videoTopLeft.y() || mev->y()<=glv->videoBottomRight.y() ) {
		xine_event_t event;
		xine_input_data_t input;
		event.type = XINE_EVENT_INPUT_MOUSE_MOVE;
		event.data = &input;
		event.data_length = sizeof(input);
		input.button = 0;
		int x, y;
		x = glv->videoBottomRight.x()-glv->videoTopLeft.x();
		y = glv->videoBottomRight.y()-glv->videoTopLeft.y();
		if ( x>0 && y>0 ) {
			input.x = (mev->x()-glv->videoTopLeft.x())*glv->videoFrame.width/x;
			input.y = (mev->y()-glv->videoTopLeft.y())*glv->videoFrame.height/y;
			xine_event_send( xine->getXineStream(), &event );
		}
	}
	mev->ignore();
}



void VideoWidget::mousePressEvent(QMouseEvent* mev)
{
	//int cur = cursor().shape();

	if ( osd->mouseClicked( mev ) )
		return;

	if ( mev->button() == Qt::LeftButton ) {
		/*if ( (cur == Qt::ArrowCursor) || (cur == Qt::BlankCursor) ) 	{
			mev->ignore();
			return;
		}*/

		if ( mev->x()>=glv->videoTopLeft.x() || mev->x()<=glv->videoBottomRight.x() || mev->y()>=glv->videoTopLeft.y() || mev->y()<=glv->videoBottomRight.y() ) {
			xine_event_t event;
			xine_input_data_t input;
			event.type = XINE_EVENT_INPUT_MOUSE_BUTTON;
			event.data = &input;
			event.data_length = sizeof(input);
			input.button = 1;
			int x, y;
			x = glv->videoBottomRight.x()-glv->videoTopLeft.x();
			y = glv->videoBottomRight.y()-glv->videoTopLeft.y();
			if ( x>0 && y>0 ) {
				input.x = (mev->x()-glv->videoTopLeft.x())*glv->videoFrame.width/x;
				input.y = (mev->y()-glv->videoTopLeft.y())*glv->videoFrame.height/y;
				xine_event_send( xine->getXineStream(), &event );
			}
		}
		mev->accept(); /* don't send event to parent */
	}
}
