/*
 * xineengine.h
 *
 * Copyright (C) 2008 Christophe Thommeret <hftom@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

#ifndef XINEENGINE_H
#define XINEENGINE_H

#include <xine.h>
#include <xine/xineutils.h>

#include <QThread>
#include <QMutex>
#include <QString>



class QWidget;
class QObject;
class QString;
class QGroupBox;
class QTextEdit;
class KPushButton;

#include <QCheckBox>

#include <KLineEdit>
#include <KNumInput>
#include <KComboBox>
#include <KDialog>

class PostFilterParameter : public QObject
{
	Q_OBJECT
public:
	PostFilterParameter(const QString &name, int offset, QWidget* parent) : QObject(parent), m_offset(offset), filterName(name) {}
	~PostFilterParameter() {};
	QString name() { return filterName; }

	virtual void setValue( const QString& ) = 0;
	virtual QString getValue() = 0;
	virtual QWidget *getWidget() = 0;

protected:
	int m_offset;
	QString filterName;
};


class PostFilterParameterInt : public PostFilterParameter
{
	Q_OBJECT
public:
	PostFilterParameterInt(const QString& name, int offset, int value, int min, int max, QWidget* parent);
	~PostFilterParameterInt() {};

	void setValue( const QString &value )
		{ int i = value.toInt(); m_numInput->setValue(i); slotIntValue(i); }
	QString getValue()
		{ QString s; s.sprintf("%d", m_numInput->value()); return s; }
	QWidget *getWidget() { return m_numInput; }

signals:
	void signalIntValue( int, int );

public slots:
	void slotIntValue(int val) { emit signalIntValue(m_offset, val); }

private:
	KIntNumInput* m_numInput;
};


class PostFilterParameterDouble : public PostFilterParameter
{
	Q_OBJECT
public:
	PostFilterParameterDouble(const QString& name, int offset, double value, double min, double max, QWidget* parent);
	~PostFilterParameterDouble() {};

	void setValue( const QString &value )
		{ double d = value.toDouble(); m_numInput->setValue(d); slotDoubleValue(d); }
	QString getValue()
		{ QString s; s.sprintf("%lf",m_numInput->value()); return s; }
	QWidget *getWidget() { return m_numInput; }

signals:
	void signalDoubleValue(int, double);

public slots:
	void slotDoubleValue(double val) { emit signalDoubleValue(m_offset, val); }

private:
	KDoubleNumInput* m_numInput;
};


class PostFilterParameterChar : public PostFilterParameter
{
	Q_OBJECT
public:
	PostFilterParameterChar(const QString& name, int offset, char *value, int size, QWidget* parent);
	~PostFilterParameterChar() {};

	void setValue(const QString &value)
		{ m_charInput->setText(value); slotCharValue(value); }
	QString getValue() { return m_charInput->text(); }
	QWidget *getWidget() { return m_charInput; }

signals:
	void signalCharValue(int, const QString&);

public slots:
	void slotCharValue(const QString& val) { emit signalCharValue(m_offset, val); }

private:
	KLineEdit* m_charInput;
};


class PostFilterParameterCombo : public PostFilterParameter
{
	Q_OBJECT
public:
	PostFilterParameterCombo(const QString& name, int offset, int value, char **enums, QWidget* parent);
	~PostFilterParameterCombo() {};

	void setValue(const QString &value)
		{ m_comboBox->setCurrentIndex(m_comboBox->findText(value)); slotIntValue(m_comboBox->currentIndex()); }
	QString getValue() { return m_comboBox->currentText(); }
	QWidget *getWidget() { return m_comboBox; }

signals:
	void signalIntValue(int, int);

public slots:
	void slotIntValue(int val) { emit signalIntValue(m_offset, val); }

private:
	KComboBox* m_comboBox;
};


class PostFilterParameterBool : public PostFilterParameter
{
	Q_OBJECT
public:
	PostFilterParameterBool(const QString& name, int offset, bool value, QWidget* parent);
	~PostFilterParameterBool() {};

	void setValue(const QString &value)
		{ bool b = (bool)value.toInt(); m_checkBox->setChecked(b); slotBoolValue(b); }
	QString getValue()
		{ QString s; s.sprintf("%d",(int)m_checkBox->isChecked()); return s; }
	QWidget *getWidget() { return m_checkBox; }

signals:
	void signalIntValue(int, int);

public slots:
	void slotBoolValue(bool val) { emit signalIntValue(m_offset, (int)val); }

private:
	QCheckBox* m_checkBox;
};


class PostFilterHelp : public KDialog
{
	Q_OBJECT
public:
	PostFilterHelp(QWidget *parent=0, const char *name=0, const QString& text = QString::null);
	~PostFilterHelp();

private:
	QTextEdit *m_textEdit;
};


class PostFilter : public QObject
{
	Q_OBJECT
public:
	PostFilter(const QString& name, xine_t* engine, xine_audio_port_t* audioDriver, xine_video_port_t* videoDriver, QWidget *parent);
	~PostFilter();

	xine_post_in_t* getInput() const;
	xine_post_out_t* getOutput() const;
	void setConfig(const QString &);
	QString getConfig();

signals:
	void signalDeleteMe( PostFilter* me );

private slots:
	void slotDeletePressed() { emit signalDeleteMe(this); }
	void slotApplyIntValue(int offset, int val);
	void slotApplyDoubleValue(int offset, double val);
	void slotApplyCharValue(int offset, const QString& val);
	void slotHelpPressed();

private:
	xine_t* m_xineEngine;
	xine_post_t* m_xinePost;
	xine_post_api_t* m_xinePostAPI;
	xine_post_api_descr_t* m_xinePostDescr;
	xine_post_api_parameter_t* m_xinePostParameter;
	char* m_data;

	QGroupBox* m_groupBox;
	QString m_filterName;

	QList<PostFilterParameter*> m_parameterList;
};



class XineEngine : public QThread
{

public:
	enum Action{ ActionPlay, ActionStop };

	XineEngine( void *parent, int formats );
	~XineEngine();
	bool play( QString mrl );
	void stop();
	xine_stream_t *getXineStream() { return stream; }
	void toggleDeinterlace();

protected:
	void run();

private:

	void unwireVideoFilters();
	void wireVideoFilters();

	xine_t *xine;
	xine_stream_t *stream;
	xine_video_port_t *voPort;
	xine_audio_port_t *aoPort;
	PostFilter* m_deinterlaceFilter;
	bool m_deinterlaceEnabled;

	QString currentMrl;
	QMutex mutex;
	Action action;
};

#endif // XINEENGINE_H
